import numpy as np
from sklearn.metrics.pairwise import euclidean_distances as EuDist2

from IDEAL_NPU import Funs
from IDEAL_NPU.cluster import EDG


X, y_true, N, dim, c_true = Funs.load_Agg()
D_full = EuDist2(X, X, squared=True)
NN_full = np.argsort(D_full, axis=1)

knn_list = [5, 6, 7, 8, 9, 10]
Y = np.zeros((len(knn_list), N))
for i, knn in enumerate(knn_list):
    NN = NN_full[:, 1:(knn+1)]
    NND = Funs.matrix_index_take(D_full, NN)

    EDG_obj = EDG(NN, NND)
    y = EDG_obj.cluster()
    Y[i, :] = y

pre = np.array([Funs.precision(y_true=y_true, y_pred=y_pred) for y_pred in Y])
rec = np.array([Funs.recall(y_true=y_true, y_pred=y_pred) for y_pred in Y])
f1 = 2 * pre * rec / (pre + rec)
ind = np.argmax(f1)

print("{}".format(pre[ind]))
print("{}".format(f1[ind]))
# print("{}".format(fmi))
# python setup.py sdist bdist_wheel




import os
import numpy as np
import scipy.io as sio
import pandas as pd
from scipy import stats
from sklearn.metrics.pairwise import euclidean_distances as EuDist2


def precision(y_true, y_pred):
    assert (len(y_pred) == len(y_true))
    N = len(y_pred)
    y_df = pd.DataFrame(data=y_pred, columns=["label"])
    ind_L = y_df.groupby("label").indices
    ni_L = [stats.mode(y_true[ind]).count[0] for yi, ind in ind_L.items()]
    return np.sum(ni_L) / N


def recall(y_true, y_pred):
    re = precision(y_true=y_pred, y_pred=y_true)
    return re


def load_Agg():
    this_directory = os.path.dirname(__file__)
    data_path = os.path.join(this_directory, "data/")
    name_full = os.path.join(data_path + "Agg.mat")
    X, y_true, N, dim, c_true = load_mat(name_full)
    return X, y_true, N, dim, c_true


def load_mat(path):
    data = sio.loadmat(path)
    X = data["X"]
    y_true = data["Y"].astype(np.int32).reshape(-1)
    N, dim, c_true = X.shape[0], X.shape[1], len(np.unique(y_true))
    return X, y_true, N, dim, c_true


def save_mat(name_full, xy):
    sio.savemat(name_full, xy)


def matrix_index_take(X, ind_M):
    assert np.all(ind_M >= 0)

    n, k = ind_M.shape
    row = np.repeat(np.array(range(n), dtype=np.int32), k)
    col = ind_M.reshape(-1)
    ret = X[row, col].reshape((n, k))
    return ret


def matrix_index_assign(X, ind_M, Val):
    n, k = ind_M.shape
    row = np.repeat(np.array(range(n), dtype=np.int32), k)
    col = ind_M.reshape(-1)
    if isinstance(Val, (float, int)):
        X[row, col] = Val
    else:
        X[row, col] = Val.reshape(-1)





import numpy as np
from sklearn.metrics.pairwise import euclidean_distances as EuDist2

from IDEAL_NPU import Funs
from IDEAL_NPU.cluster import PCN


X, y_true, N, dim, c_true = Funs.load_Agg()
D_full = EuDist2(X, X, squared=True)
NN_full = np.argsort(D_full, axis=1)

knn = 33
NN = NN_full[:, 1:(knn+1)]
NND = Funs.matrix_index_take(D_full, NN)

for i in range(N):
    tmp_ind = np.lexsort((NN[i, :], NND[i, :]))
    NN[i, :] = NN[i, tmp_ind]

print("begin")
PCN_obj = PCN(NN, NND)
y_pred = PCN_obj.cluster()
t = PCN_obj.get_time()

print("end", t)
pre = Funs.precision(y_true=y_true, y_pred=y_pred)
rec = Funs.recall(y_true=y_true, y_pred=y_pred)
f1 = 2 * pre * rec / (pre + rec)

print("{}".format(pre))
print("{}".format(f1))

# print("{}".format(pre[ind]))
# print("{}".format(f1[ind]))
# print("{}".format(fmi))



import ctypes
import os
import numpy as np


class EDG(object):
    def __init__(self, NN, NND):
        self.N, self.knn = NN.shape

        this_directory = os.path.abspath(os.path.dirname(__file__))
        self.lib = np.ctypeslib.load_library('_edg.dll', this_directory)

        self.lib.clu_new.argtypes = [ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double), ctypes.c_int32, ctypes.c_int32]
        self.lib.clu_new.restype = ctypes.c_void_p

        self.obj = self.lib.clu_new(NN.astype(np.float64).ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                               NND.astype(np.float64).ctypes.data_as(ctypes.POINTER(ctypes.c_double)), self.N, self.knn)

    def cluster(self):
        self.lib.clustering.argtypes = [ctypes.c_void_p]
        self.lib.clustering.restype = ctypes.POINTER(ctypes.c_int32)
        y_pre = self.lib.clustering(self.obj)
        y = np.ctypeslib.as_array(y_pre, (self.N,))
        return y

    def get_time(self):
        self.lib.get_time.argtypes = [ctypes.c_void_p]
        self.lib.get_time.restype = ctypes.c_double
        return self.lib.get_time(self.obj)

    def get_den(self):
        self.lib.get_den.argtypes = [ctypes.c_void_p]
        self.lib.get_den.restype = ctypes.POINTER(ctypes.c_int32)
        den = self.lib.get_den(self.obj)
        return np.ctypeslib.as_array(den, (self.N,))

    def get_row_col(self):
        self.lib.get_row_col.argtypes = [ctypes.c_void_p]
        self.lib.get_row_col.restype = ctypes.POINTER(ctypes.c_int32)
        row_col = self.lib.get_row_col(self.obj)
        nlink = row_col[0]
        rc = np.ctypeslib.as_array(row_col, (2*nlink + 1,))
        r = rc[1:(nlink+1)]
        c = rc[(nlink+1):]
        return r, c

    def get_NNS(self):
        self.lib.get_NNS.argtypes = [ctypes.c_void_p]
        self.lib.get_NNS.restype = ctypes.POINTER(ctypes.c_double)
        C_NNS = self.lib.get_NNS(self.obj)
        NNS = np.ctypeslib.as_array(C_NNS, (self.N, self.knn))
        return NNS

    @staticmethod
    def add(i, j):
        return i+j




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="IDEAL_NPU",
    version="0.4.2",
    author="Shenfei Pei",
    author_email="shenfeipei@gmail.com",
    description="A Python module for machine learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ShenfeiPei/IDEAL",
    packages=setuptools.find_packages(),
    install_requires=['numpy==1.16.5'],
    package_data={'IDEAL_NPU': ['Funs/data/Agg.mat', 'cluster/_edg.dll']},
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        'Operating System :: Microsoft :: Windows',
    ],
)



import time
import numpy as np
import numba as nb
nb.config.NUMBA_DEFAULT_NUM_THREADS = 6
import warnings
from scipy import sparse
from scipy.sparse.csgraph import connected_components


class PCN(object):
    def __init__(self, NN, NND):
        self.N, self.knn = NN.shape
        self.NN = NN
        self.NND = NND
        self.time = 0
        self.ind_sorted = np.arange(self.N*self.knn)
        self.C = np.zeros((self.N, self.knn), dtype=np.int32) - 1

    # NN n x k, NND nxk (except self)
    def sorting(self):
        aux_ind = np.arange(self.N*self.knn)
        self.ind_sorted = np.lexsort((aux_ind, self.NND.reshape(-1)))

    def compact(self):
        self.C = compact_(self.NN, self.C, self.ind_sorted)

    def compute_W(self):
        W = compute_W_(self.NN, self.C)
        return W

    def density_f(self, rho):
        density = density_f_(self.NN, self.C, rho)
        return density

    def clu_f(self, density):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            y = clu_f_(self.NN, self.C, density)
        return y

    def cluster(self):
        t_start = time.time()
        self.sorting()
        self.compact()
        W = self.compute_W()
        rho = np.sum(W, axis=1)
        density = self.density_f(rho)
        y = self.clu_f(density)
        t_end = time.time()
        self.time = t_end - t_start
        return y

    def get_time(self):
        return self.time


@nb.jit(nopython=True)
def compact_(NN, C, ind_sorted):
    N, knn = NN.shape
    for ind in ind_sorted:
        i = ind // knn
        k1 = ind % knn
        j = NN[i, k1]
        if i > j:
            if k1 == 0 or C[i, k1-1] >= 0:
                k2 = -1
                for ii in range(knn):
                    if NN[j, ii] == i:
                        k2 = ii
                        break
                if k2 >= 0:
                    if k2 == 0 or C[j, k2-1] >= 0:
                        C[i, k1] = k2
                        C[j, k2] = k1
    return C


@nb.jit(nopython=True)
def compute_Wij(NN, i, k1, j):
    N, knn = NN.shape
    sum1 = 1
    for nbi in range(k1):
        tmp_nb = NN[i, nbi]
        for ii in range(knn):
            if NN[j, ii] == tmp_nb:
                sum1 += 1
                break
    return sum1


@nb.jit(nopython=True, parallel=True)
def compute_W_(NN, C):
    N, knn = C.shape
    W = np.zeros((N, knn), dtype=np.float64)
    for i in range(N):
        for k1 in range(knn):
            if C[i, k1] >= 0:
                j = NN[i, k1]
                if j < i:
                    k2 = C[i, k1]
                    sum1 = compute_Wij(NN, i, k1, j)
                    sum2 = compute_Wij(NN, j, k2, i)
                    W[i, k1] = (sum1 + sum2) / min(k1+1, k2+1)
                    W[j, k2] = W[i, k1]
            else:
                break

    return W


@nb.jit(nopython=True)
def density_f_(NN, C, rho):
    N, knn = NN.shape
    density = np.zeros(N, dtype=np.int32)

    max_count = np.max(rho)
    rho = np.floor(rho / max_count * 100)
    thr_g = np.median(rho[rho > 0])
    for i in range(N):
        if rho[i] >= thr_g:
            density[i] = 2

    for i in range(N):
        if C[i, 0] < 0 or density[i] == 2:
            continue

        ni = 0
        for k in range(knn):
            if C[i, k] >= 0:
                ni += 1
            else:
                break

        local_rhos = rho[NN[i, :ni]]
        ind = np.where(local_rhos < thr_g)[0]
        local_rhos = local_rhos[ind]
        if len(local_rhos) <= 1:
            density[i] = 1
            continue

        thr = np.median(local_rhos)
        if rho[i] >= thr:
            density[i] = 1

    return density


@nb.jit(nopython=False)
def clu_f_(NN, C, density):
    N, knn = NN.shape
    row = list()
    col = list()
    for i in range(N):
        if density[i] == 0:
            for k1 in range(knn):
                if C[i, k1] >= 0:
                    j = NN[i, k1]
                    if density[j] > 0:
                        row.append(i)
                        col.append(j)
                        break
                else:
                    break
        else:
            for k1 in range(knn):
                if C[i, k1] >= 0:
                    j = NN[i, k1]
                    if density[j] > 0:
                        row.append(i)
                        col.append(j)
                    if density[j] == 0:
                        break
                else:
                    break

    data = list(np.ones(len(row)))
    graph = sparse.coo_matrix((data, (row, col)), shape=(N, N), copy=True)
    graph.eliminate_zeros()
    ci, y = connected_components(csgraph=graph, directed=False, return_labels=True)
    return y
